


namespace GossamerVR.AutoSave
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using UnityEngine;
    using UnityEngine.UI;
    using UnityEngine.Events;
    using UnityEngine.EventSystems;

    using SimpleJSON;    
    using Debug = UnityEngine.Debug;
    
    /// <summary>
    /// Allows saving backup files at various rates.
    /// </summary>
    public class AutoSave: MVRScript
    {
        private const string savePath = @"Saves\PluginData\";
        public const string ScriptName = "GVR AutoSave";
        private bool everythingIsTerrible = false;
        private string generalError = "";
        public bool HasError => this.everythingIsTerrible || !string.IsNullOrEmpty(generalError);

        private JSONStorableBool debugBoolJSON;
        private JSONStorableBool editModeOnlyJSON;
        private JSONStorableBool staggerSaveTimesJSON;

        private JSONStorableBool fiveMinuteJSON, fifteenMinuteJSON, thirtyMinuteJSON, lastSceneJSON;
        private JSONStorableString instructionsJSON;

        public bool IsDebugMode => this.debugBoolJSON != null && this.debugBoolJSON.val;
        public bool CanSave => !this.editModeOnlyJSON.val || SuperController.singleton.gameMode == SuperController.GameMode.Edit;
        private bool isInitialized;
        private bool isSubscribed;
        private float time;
        private float nextSaveTime;

        private List<float> lastSavedTimes = new List<float>()
        {
            0f, 0f, 0f
        };

        private readonly List<float> defaultTimes = new List<float>()
        {
            5f * 60f, 15f * 60f, 30f * 60f
        };

        private readonly List<float> staggeredTimes = new List<float>()
        {
            7f * 60f, 17f * 60f, 31f * 60f
        };

        private readonly List<string> saveNames = new List<string>()
        {
            "autosave_05m.json", "autosave_15m.json", "autosave_30m.json"
        };

        private List<bool> saveEnabled = new List<bool>()
        {
            false, true, false
        };

        /// <summary>
        /// Re-subscribes to scene events if we are re-activating post-Initialization.
        /// </summary>
        private void OnEnable()
        {            
            this.everythingIsTerrible = false;
            this.generalError = "";

            if (this.isInitialized && !this.isSubscribed)
            {                
                this.Subscribe();        
            }
        }

        /// <summary>
        /// Unsubscribes from scene events and reverts any effects we had.
        /// called automatically by Monobehaviour API when disabled.
        /// </summary>
        private void OnDisable()
        {
            this.DLog("OnDisable()");
            this.Unsubscribe();            
        }

        /// <summary>
        /// Initializes by subscribing to important scene events and 
        /// generating our json items.
        /// </summary>
        public override void Init()
        {            

            this.DLog("Init()");
            base.Init();
            if (containingAtom.name != "CoreControl" || !string.IsNullOrEmpty(containingAtom.uid))
            {
                this.Log("AutoSave must be added as a session plugin.");
                this.generalError += " invalid plugin target, please set as a session plugin !";
                this.StopWorkingAndSulk();
            }
            else
            {
                if (!this.isInitialized)
                {                    
                    this.Subscribe();
                    this.isInitialized = true;
                }
                
                this.debugBoolJSON = new JSONStorableBool("Debug Mode", false);
                this.RegisterBool(this.debugBoolJSON);

                this.editModeOnlyJSON = new JSONStorableBool("Edit Mode Only", true);
                this.RegisterBool(this.editModeOnlyJSON);

                this.staggerSaveTimesJSON = new JSONStorableBool("Stagger Save Times", true, this.OnChangeTiming);
                this.RegisterBool(this.staggerSaveTimesJSON);

                this.fiveMinuteJSON = new JSONStorableBool("5 min save", false, this.OnChangeTiming);
                this.fifteenMinuteJSON = new JSONStorableBool("15 min save", true, this.OnChangeTiming);
                this.thirtyMinuteJSON= new JSONStorableBool("30 min save", false, this.OnChangeTiming);
                //this.lastSceneJSON = new JSONStorableBool("'Last Scene' backup", true, this.OnChangeBackupSettings);

                this.instructionsJSON = new JSONStorableString("Instructions", "Each rate of autosave is saved to a different file. " +
                    "All files are stored in <b>Saves/PluginData</b> as autosave_##m.json\n\n" +
                    "<b>Edit Mode Only</b>: if true, autosave will not occur during play mode, to prevent disruption. " +
                    "if enough time passes in play mode to trigger a save, the save will occur as soon as edit mode is entered.\n" +
                    "<b>Stagger Save Times:</b> if true, files will actually save every 7, 17, and 31 minutes to reduce the number of " +
                    "simultaneous saves. This will result in more frequent saves occasionally. If false, files are saved every 5, 15, and 30 minutes, " +
                    "which results in simultaneous saves at the 15 and 30 minute mark if multiple are selected. Both frequent saving and " +
                    "simultaneous saving can be responsible for different latency-- you can have it hog a lot of resources once, or fewer resources more often." +
                    " Choose the mode that best fits your workflow."
                );

                this.CalculateNextSaveTime(this.staggeredTimes);
            }
        }

     
        /// <summary>
        /// Whenever a save mode or the stagger mode changes,
        /// we need to update the save settings and recalculate our
        /// next time;
        /// </summary>
        /// <param name="_val"></param>
        public void OnChangeTiming(bool _val)
        {
            this.saveEnabled[0] = this.fiveMinuteJSON.val;
            this.saveEnabled[1] = this.fifteenMinuteJSON.val;
            this.saveEnabled[2] = this.thirtyMinuteJSON.val;
            this.CalculateNextSaveTime(this.staggerSaveTimesJSON.val ? this.staggeredTimes : this.defaultTimes);
        }

        /// <summary>
        /// Initializes the plugin's UI, automatically called by VAM.
        /// </summary>
        public override void InitUI()
        {
            base.InitUI();
            if (this.UITransform != null && this.isInitialized)
            {
                /// LEFT
                this.CreateToggle(this.fiveMinuteJSON, false);
                this.CreateToggle(this.fifteenMinuteJSON, false);
                this.CreateToggle(this.thirtyMinuteJSON, false);
                var instructs = this.CreateTextField(this.instructionsJSON, false);
                instructs.height = 900f;

                /// RIGHT
                this.CreateToggle(this.debugBoolJSON, true);
                var spacer = this.CreateSpacer(true);
                spacer.height = 10;
                this.CreateToggle(this.editModeOnlyJSON, true);
                this.CreateToggle(this.staggerSaveTimesJSON, true);

                //this.CreateToggle(this.lastSceneJSON, true);


            }
        }

        /// <summary>
        /// Coming soon
        /// </summary>
        /// <param name="value"></param>
        private void OnChangeBackupSettings(bool value)
        {
            this.Log("Backup settings set to " + value.ToString());
        }


        // at some point we might subscribe to vam events but for now we stub it.
        private void Subscribe() => this.isSubscribed = true;
        private void Unsubscribe() => this.isSubscribed = false;

        void Update()
        {                       
            if (!this.everythingIsTerrible && this.isInitialized && (this.fiveMinuteJSON.val || this.fifteenMinuteJSON.val || this.thirtyMinuteJSON.val))
            {
                this.time += Time.unscaledDeltaTime;             
                if (this.time >= this.nextSaveTime && this.CanSave)
                {
                    this.SaveFilesByTime();
                }
            }
        }

        
        /// <summary>
        /// Iterates all 3 file settings to see if we should save,
        /// and writes the file if so.
        /// </summary>
        private void SaveFilesByTime()
        {
            List<float> thresholds = this.staggerSaveTimesJSON.val ? this.staggeredTimes : this.defaultTimes;
            List<string> saveNames = new List<string>();
            JSONClass saveData = null;
            for (int i = 0; i < thresholds.Count; i++)
            {
                if (this.saveEnabled[i] && this.time >= this.lastSavedTimes[i] + thresholds[i])
                {
                    try
                    {
                        bool wasSceneLoad = SuperController.singleton.useSceneLoadPosition;
                        saveData = saveData ?? SuperController.singleton.GetSaveJSON();
                        this.DLog($"Saving file '{this.saveNames[i]}");
                        SuperController.singleton.SaveJSON(saveData, savePath + this.saveNames[i]);
                        this.lastSavedTimes[i] = this.time;
                        SuperController.singleton.useSceneLoadPosition = wasSceneLoad;
                    }
                    catch (Exception e)
                    {
                        this.Log("Couldn't autosave: " + e.Message, error: true);
                    }
                }
            }

            this.CalculateNextSaveTime(thresholds);
        }

        /// <summary>
        /// Given a default or staggered set of times, determines 
        /// when the next earliest save should occur
        /// </summary>        
        private void CalculateNextSaveTime(List<float> timings)
        {
            this.nextSaveTime = -1f;
            float t;
            for (int i =0; i < this.lastSavedTimes.Count; i++)
            {
                if (!this.saveEnabled[i]) continue;
                t = this.lastSavedTimes[i] + timings[i];
                if (this.nextSaveTime <= this.time || t < this.nextSaveTime) this.nextSaveTime = t;
            }
        }

        /// <summary>
        /// Sets the flag showing that something is critically wrong,
        /// and disables the script. This is called when we encounter an issue
        /// that would prevent the plugin from working, such as missing transforms.        
        /// </summary>
        private void StopWorkingAndSulk()
        {
            if (!this.everythingIsTerrible)
            {

                this.Log($"{ScriptName} is shutting itself down with these errors: {this.generalError} \nThings didn't go as planned, and {ScriptName} has decided to sulk. " +
                    "You can try re-enabling or reloading it if you think it's a fluke.", error: true);

                this.everythingIsTerrible = true;
                this.generalError = "";

                // we don't want to just disable the script, we want to show it as disabled
                // in VAM's plugin UI by turning off the toggle. Though I feel like I'm doing this wrong.         
                if (this.enabledJSON != null && this.enabledJSON.toggle != null)
                {
                    this.enabledJSON.toggle.isOn = false;
                }


            }
        }

        private void DLog(string str, bool error = false)
        {
            this.LogIf(this.IsDebugMode, str, error);
        }

        private void LogIf(bool condition, string str, bool error = false)
        {
            if (condition) this.Log(str);
        }

        private void Log(string str, bool error = false)
        {
            if (error) SuperController.LogError("GVR AutoSave: "+str);
            else SuperController.LogMessage("GVR AutoSave: "+str);
        }



    }
}